// Graph/tree representation in memory: array, linked list
// Recursive/non-recursive versions of algorithm
// DFS: may cycle if loops exists (graph not tree)
// BFS: will always find path if exists, but high memory requirements
// Combination of DFS and BFS: DFS only to pre-defined depth than next node ala BFS
 
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <stdbool.h>

#include "list.h"
 
#define MAX_NODES 10
 
#define EDGE_EXISTS     1
#define EDGE_NOTEXISTS  0

/* Temporary structure for some graph traversal information */
typedef struct _item {
    int     value;              // can be used as index to array in graph_matrix
    bool bVisited;		// internal variable used to signalize that node was already visted during serach to prevent looping
    struct _item*  parent;  // parent item that cause addition of this item into saerch list (necessary to reconstruct resulting path) 
} item;

/* Matrix representation of edge-valued graph */
typedef int graph_edges[MAX_NODES][MAX_NODES];
 
/* Nice mix of edges and additional information for graph traversal,
 * within one structure 
 */ 
typedef struct _graph_matrix {
    graph_edges     edges;            // if edges[i][j] == 1 then edge between i and j exists
    item*           items[MAX_NODES]; // help array for storage of items of graph. Should hold items[i]->value == i   
} graph_matrix;
 
 
/* Set initial values for `item` */
void init_item(item* i) {
    i->value = 0;
    i->parent = NULL;
    i->bVisited = false;
}

/* Clear the auxiliary structure */
void graph_matrix_clear_visits(graph_matrix* g) {
    for (int i = 0; i < MAX_NODES; i++) {
        if  (g->items[i] != NULL) {
            g->items[i]->bVisited = false;
            g->items[i]->parent = NULL;
        }
    }
}
 
// FUNCTION FOR BOTH DEPTH- AND BREATH-FIRST SEARCH
// pathList will contain path from start 
// if (bDFS == TRUE) then depth-first search is performed, otherwise breath-first search is done 
item* graph_traversal(graph_matrix* pGraph, item* pStartItem, int targetValue, list* pathList, bool bDFS) {
    list stack;
    node *pNode;
 
    pNode = NULL;

    init_list(&stack);
    if (!pStartItem || !pGraph || !pathList) return NULL;
    // TODO:
    // 1. ADD START ITEM TO SEARCH LIST
 
    // 2. CYCLE WHILE LIST IS NOT EMPTY 
    while (pNode != NULL) { 
 
        // 3. TEST FOR TARGET ITEM
            // 4. RECONSTRUCT PATH - ONLY FOR BFS, FOR DFS IS PATH ALREADY CONSTRUCTED
	    // POINTERS FROM NODE TO ITS PARENT VIA parent ATTRIBUTE 
 
        // 5. ADD NEIGHBOURS OF REMOVED NODE INTO LIST
	// ADD ONLY UNVISITED YET
                    // 6. INSERT NODE INTO SEARCH STACK ACCORDING TO SELECTED SEARCH METHOD (DFS OR BFS)
                    // 7. SET PARENT ITEM THAT CAUSED INSERTION OF THIS ITEM INTO SEARCH LIST - NECESSARY FOR PATH RECONSTRUCTION
 
                    // 8. MARK AS VISITED
 
        // 9. GET NEXT NODE TO PROCESS FROM BEGIN OF THE STACK
    }
 
    list_release(&stack);
    return NULL;
}
 
#define START_NODE_INDEX    1
#define TARGET_NODE_INDEX   5 
#define SEARCH_REPEATS      1 
 
int main (void) {
    graph_matrix   g;
 
    srand((unsigned int) time(NULL));
    int avgLength = 0;
    int numSearches = 0; 
 
    // REPEAT WHOLE SEARCH SEVERAL TIMES WITH DIFFERENT EDGE MATRICES
    for (int k = 0; k < SEARCH_REPEATS; k++) {
        // CREATE NODES OF GRAPH
        for (int i = 0; i < MAX_NODES; i++) {
            g.items[i] = (item*) malloc(sizeof(item));
            init_item(g.items[i]);
            g.items[i]->value = i;
        }
 
        // CLEAR EDGES    
        for (int i = 0; i < MAX_NODES; i++) {
            for (int j = 0; j < MAX_NODES; j++) {
                g.edges[i][j] = 0;
            }
        }
 
        // CREATE RANDOM GRAPH EDGES
        for (int i = 0; i < MAX_NODES; i++) {
            for (int j = 0; j < MAX_NODES; j++) {
                // GENERATE RANDOM EDGE - 0 == no edge, 1 == edge from i to j
                g.edges[i][j] = rand() % 2;
                // NO EDGE TO NODE ITSELF
                if (i == j) g.edges[i][j] = 0;
            }
        }
 
 
        // PRINT EDGES MATRIX
        printf("   ");
        for (int i = 0; i < MAX_NODES; i++) printf("%d ", i);
        printf("\n");
        for (int i = 0; i < MAX_NODES; i++) {
            printf("%d: ", i);
            for (int j = 0; j < MAX_NODES; j++) {
                printf("%d ", g.edges[i][j]);
            }
            printf("\n");
        }
        printf("\n");
 
        // START WITH NODE START_NODE_INDEX
        // SEARCH FOR PATH TO NODE TARGET_NODE_INDEX
        list path_list;
        init_list(&path_list);

        if (graph_traversal(&g, g.items[START_NODE_INDEX], TARGET_NODE_INDEX, &path_list, true) != NULL) {
            numSearches++;
            // PRINT PATH
            printf("PATH:");
            node* pNode = path_list.begin;
            while (pNode != NULL) {
               printf(" -> %d", ((item*)pNode->data)->value);
               pNode = pNode->next;
               avgLength++;
            }
            avgLength--;    // REMOVE START NODE FROM COUNTED PATH LENGTH
            printf("\n\n");
        }
        else {
            printf("No path found!");
        }
 
        // RELEASE MEMORY
        list_release(&path_list);
        for (int i = 0; i < MAX_NODES; i++) {
            free(g.items[i]);
        }
    }
 
    printf("Avg path length: %f\n",  (avgLength / (float) numSearches));

    return 0;
}

